using System;
using Gtk;
using PathWords;

public partial class MainWindow: Gtk.Window
{	
#region Enumerations
	/**
	 * Potention board sizes for the game
	 */
	public enum BoardSize {
		SMALL,
		MEDIUM,
		LARGE
	};
#endregion
#region Data Members
	/**
	 * Number of columns on the board
	 */
	public static readonly int[] Columns = new int[]{
		7, /* SMALL  */
		8, /* MEDIUM */
		9  /* LARGE  */
	};

	/**
	 * Default size of the board
	 */
	protected const BoardSize DEFAULT_SIZE = BoardSize.MEDIUM;

	/**
	 * Current size of the board
	 */
	private int boardSize = (int)DEFAULT_SIZE;

	/**
	 * Number of rows in an even column
	 */
	public const int ROWS_EVEN = 6;

	/**
	 * Number of rows in an odd column
	 */
	public const int ROWS_ODD  = 7;

	/**
	 * Maximum number of rows (only for initializing arrays)
	 */
	private const int MAX_ROWS = ROWS_ODD;
    
	/**
	 * Action item for the New Game menu button
	 */
	private Gtk.Action NewGameAction;

	/**
	 * Action item for the Small Board size menu button
	 */
	private RadioAction SmallBoardAction;

	/**
	 * Action item for the Medium Board size menu button
	 */
	private RadioAction MediumBoardAction;

	/**
	 * Action item for the Large Board size menu button
	 */
	private RadioAction LargeBoardAction;

	/**
	 * Action item for the Quit menu button
	 */
	private Gtk.Action QuitAction;

	/**
	 * Array of entry cells
	 */
	private PathWords.NeighborEntry[,] Entries;

	/**
	 * Layout that holds all of the entries
	 */
	Gtk.Fixed fixedLayout;

	/**
	 * Default Button
	 */
	private Gtk.Button executeButton;
    
	/**
	 * Cancel button
	 */
	private Gtk.Button cancelButton;
	
	/**
	 * Array of path tiles, which are represented by entry cells
	 */
	private PathTile[,] Tiles;

	/**
	 * List of possible words to use
	 */
	private WordDictionary dictionary;

	/**
	 * Current best path found
	 */
	private WordPath BestPath = null;

	/**
	 * Queue of entry cells which should be filled in next
	 */
	private System.Collections.Generic.List<NeighborEntry> NextQueue;
#endregion
	
#region Constructors
	public MainWindow (): base (Gtk.WindowType.Toplevel)
	{
		InitUI();
		InitData();
		InitHandlers();

		InitNewGame();
	}
#endregion

#region Member Functions
#region Initialization Methods
	protected void InitUI() {
		// MainWindow
		Title = "PathWords";
		Icon = Gdk.Pixbuf.LoadFromResource("pathwords-x.png");
		Resizable = false;
		
		// VBox
		Gtk.VBox vbox = new Gtk.VBox();
		vbox.Spacing = 5;

		// Menu Bar
		Gtk.UIManager manager = new Gtk.UIManager();
		Gtk.ActionGroup menuAction = new Gtk.ActionGroup("Default");
		
		// File Menu
		Gtk.Action fileMenu = new Gtk.Action("FileAction", "_File");
		fileMenu.ShortLabel = "_File";
		menuAction.Add(fileMenu, null);

		NewGameAction = new Gtk.Action("NewGameAction", "_New Game",
				null, "gtk-new");
		NewGameAction.ShortLabel = "_New Game";
		menuAction.Add(NewGameAction, null);

		Action boardSizeAction = new Action("BoardSizeAction", "_Board Size");
		boardSizeAction.ShortLabel = "_Board Size";
		menuAction.Add(boardSizeAction, null);

		SmallBoardAction = new RadioAction("SmallBoardAction", "_Small", null, null, (int)BoardSize.SMALL);
		SmallBoardAction.Label = "_Small";
		// SmallBoardAction.Active = (BoardSize.SMALL == DEFAULT_SIZE);
		SmallBoardAction.Group = new GLib.SList(System.IntPtr.Zero);;
		menuAction.Add(SmallBoardAction, null);

		MediumBoardAction = new RadioAction("MediumBoardAction", "_Medium", null, null, (int)BoardSize.MEDIUM);
		MediumBoardAction.Label = "_Medium";
		// MediumBoardAction.Active = (BoardSize.MEDIUM == DEFAULT_SIZE);
		MediumBoardAction.Group = SmallBoardAction.Group;
		menuAction.Add(MediumBoardAction, null);

		LargeBoardAction = new RadioAction("LargeBoardAction", "_Large", null, null, (int)BoardSize.LARGE);
		LargeBoardAction.Label = "_Large";
		// LargeBoardAction.Active = (BoardSize.LARGE == DEFAULT_SIZE);
		LargeBoardAction.Group = SmallBoardAction.Group;
		menuAction.Add(LargeBoardAction, null);

		QuitAction = new Gtk.Action("QuitAction", "_Quit",
				null, "gtk-quit");
		QuitAction.ShortLabel = "_Quit";
		menuAction.Add(QuitAction, null);

		// More Menu Bar
		manager.InsertActionGroup(menuAction, 0);
		AddAccelGroup(manager.AccelGroup);
		manager.AddUiFromString("<ui><menubar name='menuBar'><menu action='FileAction'><menuitem action='NewGameAction'/><menu action='BoardSizeAction'><menuitem action='SmallBoardAction'/><menuitem action='MediumBoardAction'/><menuitem action='LargeBoardAction'/></menu><menuitem action='QuitAction'/></menu></menubar></ui>");
		Gtk.MenuBar menuBar = ((Gtk.MenuBar)(manager.GetWidget("/menuBar")));
		vbox.Add(menuBar);
		Gtk.Box.BoxChild boxChild = ((Gtk.Box.BoxChild)(vbox[menuBar]));
		boxChild.Position = 0;
		boxChild.Expand = false;
		boxChild.Fill = false;

		// fixedLayout
		fixedLayout = new Gtk.Fixed();
		fixedLayout.HasWindow = false;
		vbox.Add(fixedLayout);
		boxChild = ((Gtk.Box.BoxChild)(vbox[fixedLayout]));
		boxChild.Position = 1;
		boxChild.Expand = true;
		boxChild.Fill = true;

		// Entries
		InitEntries();
		
		// HBox
		HBox hbox = new HBox();
		hbox.Spacing = 5;

		// executeButton
		executeButton = new Gtk.Button();
		executeButton.CanDefault = true;
		executeButton.CanFocus = true;
		executeButton.Label = "Execute";
		hbox.Add(executeButton);
		boxChild = ((Box.BoxChild)(hbox[executeButton]));
		boxChild.Position = 0;
		
		// cancelButton
		cancelButton = new Gtk.Button();
		cancelButton.Sensitive = false;
		cancelButton.Label = "Cancel";
		hbox.Add(cancelButton);
		boxChild = ((Box.BoxChild)(hbox[cancelButton]));
		boxChild.Position = 1;
		boxChild.Expand = false;
		boxChild.Fill = false;

		vbox.Add(hbox);
		boxChild = ((Box.BoxChild)(vbox[hbox]));
		boxChild.Position = 2;
		boxChild.Expand = false;
		boxChild.Fill = false;
		
		// Pack it in
		Add(vbox);
		executeButton.HasDefault = true;
		MediumBoardAction.Active = true;

		if ((Child != null)) {
		    Child.ShowAll();
		}
		Show();
	}

	private void InitEntries() {

		if (Entries != null)
			foreach(NeighborEntry entry in Entries)
				if (entry != null)
					fixedLayout.Remove(entry);

		Entries = new PathWords.NeighborEntry[
			Columns[boardSize],MAX_ROWS];
		for (int i = 0; i < Columns[boardSize]; i++) {
			for (int j = 0; j < (i%2==0 ? ROWS_EVEN : ROWS_ODD); j++) {
				Entries[i,j] = new NeighborEntry();
				Entries[i,j].CanFocus = true;
				Entries[i,j].IsEditable = true;
				Entries[i,j].WidthChars = 2;
				Entries[i,j].MaxLength = 2;
				Entries[i,j].ActivatesDefault = true;
				
				fixedLayout.Add(Entries[i,j]);
				Gtk.Fixed.FixedChild fc = ((Gtk.Fixed.FixedChild)(fixedLayout[Entries[i,j]]));
				fc.X = 30*i;
				fc.Y = 30*j + (i%2==0 ? 15 : 0);
			}
		}
		// Separate loop so we know entries aren't null
		for (int i = 0; i < Columns[boardSize]; i++)
			for (int j = 0; j < (i%2==0 ? ROWS_EVEN : ROWS_ODD); j++)
				InitEntryNeighbors(i,j);

	}

	private void InitEntryNeighbors(int x, int y) {
		int rows = (x%2==0 ? ROWS_EVEN : ROWS_ODD);
		int oppRows = (x%2==0 ? ROWS_ODD : ROWS_EVEN);
		int rowsInLastCol = (Columns[boardSize]%2==0 ?
				ROWS_ODD : ROWS_EVEN);
		int lastCellCol = (Columns[boardSize]%2==0 ? 
				Columns[boardSize]-1 : 
				Columns[boardSize]-2);

		if (x == 0 && y == 0) {
			Entries[x,y].RightNeighbor = Entries[x+1,y];
			Entries[x,y].LeftNeighbor  = Entries[lastCellCol, ROWS_ODD-1];
		} else if (x == Columns[boardSize]-1 && y == rowsInLastCol-2) {
			Entries[x,y].RightNeighbor = Entries[1, ROWS_ODD-1];
			Entries[x,y].LeftNeighbor  = Entries[Columns[boardSize]-2, y];
		} else if (x == 1 && y == ROWS_ODD-1) {
			Entries[x,y].RightNeighbor = Entries[x+2, y];
			Entries[x,y].LeftNeighbor  = Entries[Columns[boardSize]-1, rowsInLastCol-2];
		} else if (x == lastCellCol && y == ROWS_ODD-1) {
			Entries[x,y].RightNeighbor = Entries[0,0];
			Entries[x,y].LeftNeighbor  = Entries[x-2, y];
		} else if (y == ROWS_ODD-1) {
			Entries[x,y].RightNeighbor = Entries[x+2, y];
			Entries[x,y].LeftNeighbor  = Entries[x-2, y];
		} else {
			Entries[x,y].LeftNeighbor   = Entries[
				(x+Columns[boardSize]-1)%Columns[boardSize], (x-1 < 0) ? (y+rows-1)%rows : y ];
			Entries[x,y].RightNeighbor  = Entries[
				(x+1)%Columns[boardSize], (x+1 > Columns[boardSize]-1) ? (y+1)%rows : y ];
		}
		Entries[x,y].TopNeighbor    = Entries[
			(y-1 < 0) ? (x+Columns[boardSize]-1)%Columns[boardSize] : x, (y+oppRows-1)%oppRows ];
		Entries[x,y].BottomNeighbor = Entries[
			(y+1 > rows-1) ? (x+1)%Columns[boardSize] : x, (y+1)%rows ];

	}

	private void InitTileNeighbors(int x, int y) {
		int rows = (x%2==0 ? ROWS_EVEN : ROWS_ODD);

		System.Collections.Generic.List<PathTile> neighbors =
			new System.Collections.Generic.List<PathTile>();

		if (x>0 && y<ROWS_ODD-1) neighbors.Add(Tiles[x-1,y]);
		if (x<Columns[boardSize]-1 && y<ROWS_ODD-1) neighbors.Add(Tiles[x+1,y]);
		if (y>0) neighbors.Add(Tiles[x,y-1]);
		if (y<rows-1) neighbors.Add(Tiles[x,y+1]);
		if (x%2 == 1 && y>0) {
			if (x>0) neighbors.Add(Tiles[x-1, y-1]);
			if (x<Columns[boardSize]-1) neighbors.Add(Tiles[x+1, y-1]);
		} else if (x%2==0) {
			if (x>0) neighbors.Add(Tiles[x-1, y+1]);
			if (x<Columns[boardSize]-1) neighbors.Add(Tiles[x+1, y+1]);
		}

		Tiles[x,y].Neighbors = neighbors.ToArray();
	}
	
	protected void InitHandlers() {
		// Global Events
		DeleteEvent += new Gtk.DeleteEventHandler(OnDeleteEvent);

		// Menu Buttons
		NewGameAction.Activated += new System.EventHandler(OnNewGameActivated);
		SmallBoardAction.Activated += new System.EventHandler(OnBoardSizeActivated);
		MediumBoardAction.Activated += new System.EventHandler(OnBoardSizeActivated);
		LargeBoardAction.Activated += new System.EventHandler(OnBoardSizeActivated);
		QuitAction.Activated += new System.EventHandler(OnQuitActivated);
		
		// Fixed Layout
		fixedLayout.ExposeEvent += new ExposeEventHandler(OnExposeEvent);

		// Entries
		InitEntryHandlers();

		// Execute Button
		this.executeButton.Clicked += new System.EventHandler(this.OnExecuteButtonClicked);
		
		// Cancel Button
		this.cancelButton.Clicked += new System.EventHandler(this.OnCancelButtonClicked);
	}

	protected void InitData() {
		//Dictionary
		dictionary = new WordDictionary(Globals.DATA_DIR + "/Dictionary.txt");

		// NextQueue
		NextQueue = new System.Collections.Generic.List<NeighborEntry>();

		// Tiles
		InitTiles();
	}

	private void InitTiles() {

		// Tiles
		Tiles = new PathTile[Columns[boardSize],MAX_ROWS];
		for (int i = 0; i < Columns[boardSize]; i++)
			for (int j = 0; j < (i%2==0 ? ROWS_EVEN : ROWS_ODD); j++)
				Tiles[i,j] = new PathTile(Entries[i,j]);
		// Separate loop so we know entries aren't null
		for (int i = 0; i < Columns[boardSize]; i++)
			for (int j = 0; j < (i%2==0 ? ROWS_EVEN : ROWS_ODD); j++) {
				InitTileNeighbors(i,j);
				NextQueue.Add(Entries[i,j]);
			}
	}

	private void InitEntryHandlers() {

		for (int i = 0; i < Columns[boardSize]; i++) {
			for (int j = 0; j < (i%2==0 ? ROWS_EVEN : ROWS_ODD); j++) {
				// FIXME: Move these to outside function calls
				Entries[i,j].KeyPressEvent += delegate(object o, KeyPressEventArgs args) {
					Gdk.Key key = args.Event.Key;
					if ((key == Gdk.Key.Right || key == Gdk.Key.Tab))
						FocusNextNeighbor((NeighborEntry)o);
				};
				Entries[i,j].TextInserted += delegate(object o, TextInsertedArgs args) {
					if (args.Text.Length > 0) {
						NeighborEntry entry = (NeighborEntry)o;
						
						if (NextQueue.Contains(entry)) {
						    NextQueue.Remove(entry);
						}
						
						FocusNextNeighbor((NeighborEntry)o);
					}
				};
				Entries[i,j].FocusGrabbed += delegate(object sender, EventArgs e) {
					((NeighborEntry)sender).SelectRegion(0,-1);
				};
			}
		}
	}

	protected void InitNewGame() {

		executeButton.Label = "Execute";
		BestPath = null;
		fixedLayout.QueueDraw();
		NextQueue.Clear();

		for (int i = 0; i < Columns[boardSize]; i++)
			for (int j = 0; j < (i%2==0 ? ROWS_EVEN : ROWS_ODD); j++) {
				Entries[i,j].Text = string.Empty;
				NextQueue.Add(Entries[i,j]);
			}
		Entries[0,0].GrabFocus();
	}
#endregion

#region Event Handlers
	protected void OnExposeEvent(object o, ExposeEventArgs args) {
		if (BestPath != null) {

			Widget widget = o as Widget;
			Gdk.Point[] points = new Gdk.Point[
				BestPath.TilePath.Count];
			for (int i = 0; i < points.Length; i++) {
				
				Gdk.Rectangle rect = BestPath.TilePath[i].Entry.Allocation;
				points[i] = new Gdk.Point(rect.X + rect.Width/2,
						rect.Y + rect.Height/2);
			}

			GdkWindow.DrawLines(Style.ForegroundGC(State), points);
		}
	}

	protected void OnDeleteEvent (object sender, DeleteEventArgs a)
	{
		Application.Quit ();
		a.RetVal = true;
	}

	protected virtual void OnNewGameActivated (object sender, System.EventArgs e) {
		InitNewGame();
	}

	protected virtual void OnBoardSizeActivated (object sender, System.EventArgs e) {
		int newSize = ((RadioAction)sender).CurrentValue;
		if (newSize != boardSize) {
			boardSize = newSize;

			InitEntries();
			InitTiles();
			InitEntryHandlers();

			if ((Child != null)) {
			    Child.ShowAll();
			}
			Show();

			InitNewGame();
		}
	}

	protected virtual void OnQuitActivated (object sender, System.EventArgs e) {
		Application.Quit ();
	}

	protected virtual void OnExecuteButtonClicked (object sender, System.EventArgs e) {
		
		if (executeButton.Label.Equals("Execute"))
			FindBestWord();
		else // Update
			UpdateTiles();
	}

	protected virtual void OnCancelButtonClicked (object sender, System.EventArgs e) {
		
		BestPath = null;
		executeButton.Label = "Execute";
		cancelButton.Sensitive = false;
		fixedLayout.QueueDraw();
	}
#endregion
	
	private void FocusNextNeighbor(NeighborEntry entry) {
		
		if (NextQueue.Count != 0) {
			NextQueue[0].GrabFocus();
		} else {
			// Simply select the current cell
			entry.SelectRegion(0,-1);
		}
	}

	
	protected void UpdateTiles() {

		foreach(PathTile tile in BestPath.TilePath)
			tile.Entry.Text = string.Empty;
	
		for (int i = 0; i < Columns[boardSize]; i++) {
			for (int j = (i%2==0 ? ROWS_EVEN : ROWS_ODD)-1; j > 0; j--) {
				if (Entries[i,j].Text.Equals(string.Empty)) {
					int idx = j-1;
					while (idx >= 0 && Entries[i,idx].Text.Equals(string.Empty))
						idx--;
					if (idx < 0) break;

					Entries[i,j].Text = Entries[i,idx].Text;
					Entries[i,idx].Text = string.Empty;
				}
			}
		}
		
		for (int i = 0; i < Columns[boardSize]; i++) {
			for (int j = 0; j < (i%2==0 ? ROWS_EVEN : ROWS_ODD); j++)
				if (Entries[i,j].Text.Equals(string.Empty) && !NextQueue.Contains(Entries[i,j]))
					NextQueue.Add(Entries[i,j]);
		}
		
		if (NextQueue.Count != 0)
			NextQueue[0].GrabFocus();
		
		BestPath = null;
		executeButton.Label = "Execute";
		cancelButton.Sensitive = false;
		fixedLayout.QueueDraw();
	}
	
	protected void FindBestWord() {
		BestPath = null;

		try {
		
			WordPath maxPath = null;
			int maxLength = -1;
			
			for (int i = 0; i < Columns[boardSize]; i++)
			for (int j = 0; j < (i%2==0 ? ROWS_EVEN : ROWS_ODD); j++) {
				PathTile curTile = Tiles[i,j];
				
				WordPath curPath = new WordPath(dictionary, curTile).FindLongestPath();
				if (WordPath.PathLength(curPath) > maxLength) {
					maxPath = curPath;
					maxLength = WordPath.PathLength(maxPath);
				}
				
			}
			
		 Console.WriteLine("Best overall: {0}", maxPath.currentWord);
			for (int i = 0; i < Columns[boardSize]; i++) {
				for (int j = 0; j < (i%2==0 ? ROWS_EVEN : ROWS_ODD); j++) {
					Entries[i,j].SelectRegion(0,0);
				}
			}

			maxPath.TilePath[0].Entry.GrabFocus();
			executeButton.Label = maxPath.currentWord;
			cancelButton.Sensitive = true;
			BestPath = maxPath;
			fixedLayout.QueueDraw();
			
		} catch (EmptyTileException ex) {
		
			if(!NextQueue.Contains(ex.Tile.Entry))
				NextQueue.Add(ex.Tile.Entry);
			
			ex.Tile.Entry.GrabFocus();
		}
	}
}
#endregion
